import pytest
from flask import Flask, url_for
from flask_sqlalchemy import SQLAlchemy
from marshmallow import ValidationError
from werkzeug.wrappers import Response

from flask_marshmallow import Marshmallow
from flask_marshmallow.sqla import HyperlinkRelated
from tests.conftest import Bunch

try:
    from marshmallow_sqlalchemy import SQLAlchemySchema  # noqa: F401
except ImportError:
    has_sqlalchemyschema = False
else:
    has_sqlalchemyschema = True


requires_sqlalchemyschema = pytest.mark.skipif(
    not has_sqlalchemyschema, reason="SQLAlchemySchema not available"
)


class TestSQLAlchemy:
    @pytest.fixture
    def extapp(self):
        app_ = Flask("extapp")
        app_.testing = True
        app_.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
        app_.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        SQLAlchemy(app_)
        Marshmallow(app_)

        @app_.route("/author/<int:id>")
        def author(id):
            return f"...view for author {id}..."

        @app_.route("/book/<int:id>")
        def book(id):
            return f"...view for book {id}..."

        ctx = app_.test_request_context()
        ctx.push()

        yield app_

        ctx.pop()

    @pytest.fixture
    def db(self, extapp):
        db = extapp.extensions["sqlalchemy"]
        yield db
        db.session.close()
        db.engine.dispose()

    @pytest.fixture
    def extma(self, extapp):
        return extapp.extensions["flask-marshmallow"]

    @pytest.fixture
    def models(self, db):
        class AuthorModel(db.Model):
            __tablename__ = "author"
            id = db.Column(db.Integer, primary_key=True)
            name = db.Column(db.String(255))

            @property
            def url(self):
                return url_for("author", id=self.id)

            @property
            def absolute_url(self):
                return url_for("author", id=self.id, _external=True)

        class BookModel(db.Model):
            __tablename__ = "book"
            id = db.Column(db.Integer, primary_key=True)
            title = db.Column(db.String(255))
            author_id = db.Column(db.Integer, db.ForeignKey("author.id"))
            author = db.relationship("AuthorModel", backref="books")

            @property
            def url(self):
                return url_for("book", id=self.id)

            @property
            def absolute_url(self):
                return url_for("book", id=self.id, _external=True)

        db.create_all()
        yield Bunch(Author=AuthorModel, Book=BookModel)
        db.drop_all()

    def test_can_initialize_extensions(self, extapp):
        assert "flask-marshmallow" in extapp.extensions
        assert "sqlalchemy" in extapp.extensions

    @requires_sqlalchemyschema
    def test_can_declare_sqla_schemas(self, extma, models, db):
        class AuthorSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Author

            id = extma.auto_field()
            name = extma.auto_field()

        class BookSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Book

            id = extma.auto_field()
            title = extma.auto_field()
            author_id = extma.auto_field()

        author_schema = AuthorSchema()
        book_schema = BookSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)

        author_result = author_schema.dump(author)

        assert "id" in author_result
        assert "name" in author_result
        assert author_result["id"] == author.id
        assert author_result["name"] == "Chuck Paluhniuk"
        book_result = book_schema.dump(book)

        assert "id" in book_result
        assert "title" in book_result
        assert book_result["id"] == book.id
        assert book_result["title"] == book.title
        assert book_result["author_id"] == book.author_id

        resp = author_schema.jsonify(author)
        assert isinstance(resp, Response)

    @requires_sqlalchemyschema
    def test_can_declare_sqla_auto_schemas(self, extma, models, db):
        class AuthorSchema(extma.SQLAlchemyAutoSchema):
            class Meta:
                model = models.Author

        class BookSchema(extma.SQLAlchemyAutoSchema):
            class Meta:
                model = models.Book
                include_fk = True

            id = extma.auto_field()
            title = extma.auto_field()
            author_id = extma.auto_field()

        author_schema = AuthorSchema()
        book_schema = BookSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)

        author_result = author_schema.dump(author)

        assert "id" in author_result
        assert "name" in author_result
        assert author_result["id"] == author.id
        assert author_result["name"] == "Chuck Paluhniuk"
        book_result = book_schema.dump(book)

        assert "id" in book_result
        assert "title" in book_result
        assert book_result["id"] == book.id
        assert book_result["title"] == book.title
        assert book_result["author_id"] == book.author_id

        resp = author_schema.jsonify(author)
        assert isinstance(resp, Response)

    # FIXME: temporarily filter out this warning
    # this is triggered by marshmallow-sqlalchemy on sqlalchemy v1.4.x
    # on the current version it should be fixed
    # in an upcoming marshmallow-sqlalchemy release
    @requires_sqlalchemyschema
    def test_hyperlink_related_field(self, extma, models, db, extapp):
        class BookSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Book

            author = extma.HyperlinkRelated("author")

        book_schema = BookSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)
        db.session.add(author)
        db.session.add(book)
        db.session.flush()

        book_result = book_schema.dump(book)

        assert book_result["author"] == author.url

        deserialized = book_schema.load(book_result)
        assert deserialized["author"] == author

    @requires_sqlalchemyschema
    def test_hyperlink_related_field_serializes_none(self, extma, models):
        class BookSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Book

            author = extma.HyperlinkRelated("author")

        book_schema = BookSchema()
        book = models.Book(title="Fight Club", author=None)
        book_result = book_schema.dump(book)
        assert book_result["author"] is None

    @requires_sqlalchemyschema
    def test_hyperlink_related_field_errors(self, extma, models, db, extapp):
        class BookSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Book

            author = HyperlinkRelated("author")

        book_schema = BookSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)
        db.session.add(author)
        db.session.add(book)
        db.session.flush()

        # Deserialization fails on bad endpoint
        book_result = book_schema.dump(book)
        book_result["author"] = book.url
        with pytest.raises(ValidationError) as excinfo:
            book_schema.load(book_result)
        errors = excinfo.value.messages
        assert 'expected "author"' in errors["author"][0]

        # Deserialization fails on bad URL key
        book_result = book_schema.dump(book)
        book_schema.fields["author"].url_key = "pk"
        with pytest.raises(ValidationError) as excinfo:
            book_schema.load(book_result)
        errors = excinfo.value.messages
        assert 'URL pattern "pk" not found' in errors["author"][0]

    @requires_sqlalchemyschema
    def test_hyperlink_related_field_external(self, extma, models, db, extapp):
        class BookSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Book

            author = HyperlinkRelated("author", external=True)

        book_schema = BookSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)
        db.session.add(author)
        db.session.add(book)
        db.session.flush()

        book_result = book_schema.dump(book)

        assert book_result["author"] == author.absolute_url

        deserialized = book_schema.load(book_result)
        assert deserialized["author"] == author

    @requires_sqlalchemyschema
    def test_hyperlink_related_field_list(self, extma, models, db, extapp):
        class AuthorSchema(extma.SQLAlchemySchema):
            class Meta:
                model = models.Author

            books = extma.List(HyperlinkRelated("book"))

        author_schema = AuthorSchema()

        author = models.Author(name="Chuck Paluhniuk")
        book = models.Book(title="Fight Club", author=author)
        db.session.add(author)
        db.session.add(book)
        db.session.flush()

        author_result = author_schema.dump(author)
        assert author_result["books"][0] == book.url

        deserialized = author_schema.load(author_result)
        assert deserialized["books"][0] == book



import io
from tempfile import SpooledTemporaryFile

import pytest
from flask import url_for
from marshmallow import missing
from marshmallow.exceptions import ValidationError
from werkzeug.datastructures import FileStorage
from werkzeug.routing import BuildError

from flask_marshmallow.fields import _tpl


@pytest.mark.parametrize(
    "template", ["<id>", " <id>", "<id> ", "< id>", "<id  >", "< id >"]
)
def test_tpl(template):
    assert _tpl(template) == "id"
    assert _tpl(template) == "id"
    assert _tpl(template) == "id"


def test_url_field(ma, mockauthor):
    field = ma.URLFor("author", values=dict(id="<id>"))
    result = field.serialize("url", mockauthor)
    assert result == url_for("author", id=mockauthor.id)

    mockauthor.id = 0
    result = field.serialize("url", mockauthor)
    assert result == url_for("author", id=0)


def test_url_field_with_invalid_attribute(ma, mockauthor):
    field = ma.URLFor("author", values=dict(id="<not-an-attr>"))
    expected_msg = "{!r} is not a valid attribute of {!r}".format(
        "not-an-attr", mockauthor
    )
    with pytest.raises(AttributeError, match=expected_msg):
        field.serialize("url", mockauthor)


def test_url_field_handles_nested_attribute(ma, mockbook, mockauthor):
    field = ma.URLFor("author", values=dict(id="<author.id>"))
    result = field.serialize("url", mockbook)
    assert result == url_for("author", id=mockauthor.id)


def test_url_field_handles_none_attribute(ma, mockbook, mockauthor):
    mockbook.author = None

    field = ma.URLFor("author", values=dict(id="<author>"))
    result = field.serialize("url", mockbook)
    assert result is None

    field = ma.URLFor("author", values=dict(id="<author.id>"))
    result = field.serialize("url", mockbook)
    assert result is None


def test_url_field_deserialization(ma):
    field = ma.URLFor("author", values=dict(id="<not-an-attr>"), allow_none=True)
    # noop
    assert field.deserialize("foo") == "foo"
    assert field.deserialize(None) is None


def test_invalid_endpoint_raises_build_error(ma, mockauthor):
    field = ma.URLFor("badendpoint")
    with pytest.raises(BuildError):
        field.serialize("url", mockauthor)


def test_hyperlinks_field(ma, mockauthor):
    field = ma.Hyperlinks(
        {
            "self": ma.URLFor("author", values={"id": "<id>"}),
            "collection": ma.URLFor("authors"),
        }
    )

    result = field.serialize("_links", mockauthor)
    assert result == {
        "self": url_for("author", id=mockauthor.id),
        "collection": url_for("authors"),
    }


def test_hyperlinks_field_recurses(ma, mockauthor):
    field = ma.Hyperlinks(
        {
            "self": {
                "href": ma.URLFor("author", values={"id": "<id>"}),
                "title": "The author",
            },
            "collection": {"href": ma.URLFor("authors"), "title": "Authors list"},
        }
    )
    result = field.serialize("_links", mockauthor)

    assert result == {
        "self": {"href": url_for("author", id=mockauthor.id), "title": "The author"},
        "collection": {"href": url_for("authors"), "title": "Authors list"},
    }


def test_hyperlinks_field_recurses_into_list(ma, mockauthor):
    field = ma.Hyperlinks(
        [
            {"rel": "self", "href": ma.URLFor("author", values={"id": "<id>"})},
            {"rel": "collection", "href": ma.URLFor("authors")},
        ]
    )
    result = field.serialize("_links", mockauthor)

    assert result == [
        {"rel": "self", "href": url_for("author", id=mockauthor.id)},
        {"rel": "collection", "href": url_for("authors")},
    ]


def test_hyperlinks_field_deserialization(ma):
    field = ma.Hyperlinks(
        {"href": ma.URLFor("author", values={"id": "<id>"})}, allow_none=True
    )
    # noop
    assert field.deserialize("/author") == "/author"
    assert field.deserialize(None) is None


def test_absolute_url(ma, mockauthor):
    field = ma.AbsoluteURLFor("authors")
    result = field.serialize("abs_url", mockauthor)
    assert result == url_for("authors", _external=True)


def test_absolute_url_deserialization(ma):
    field = ma.AbsoluteURLFor("authors", allow_none=True)
    assert field.deserialize("foo") == "foo"
    assert field.deserialize(None) is None


def test_aliases(ma):
    from flask_marshmallow.fields import AbsoluteURLFor, AbsoluteUrlFor, URLFor, UrlFor

    assert UrlFor is URLFor
    assert AbsoluteUrlFor is AbsoluteURLFor


def test_file_field(ma, mockauthor):
    field = ma.File()
    fs = FileStorage(io.BytesIO(b"test"), "test.jpg")
    result = field.deserialize(fs, mockauthor)
    assert result == fs

    with SpooledTemporaryFile() as temp:
        temp.write(b"temp")
        fs = FileStorage(temp, "temp.jpg")
        result = field.deserialize(fs, mockauthor)
        assert result == fs

    result = field.deserialize("", mockauthor)
    assert result is missing

    with pytest.raises(ValidationError, match="Field may not be null."):
        field.deserialize(None, mockauthor)

    with pytest.raises(ValidationError, match="Not a valid file."):
        field.deserialize("123", mockauthor)


def test_config_field(ma, app, mockauthor):
    app.config["NAME"] = "test"
    field = ma.Config(key="NAME")

    result = field.serialize("config_value", mockauthor)
    assert result == "test"

    field = ma.Config(key="DOES_NOT_EXIST")
    with pytest.raises(ValueError, match="not found in the app config"):
        field.serialize("config_value", mockauthor)



import io
from tempfile import SpooledTemporaryFile

import pytest
from marshmallow.exceptions import ValidationError
from werkzeug.datastructures import FileStorage

from flask_marshmallow import validate


@pytest.mark.parametrize("size", ["1 KB", "1 KiB", "1 MB", "1 MiB", "1 GB", "1 GiB"])
def test_parse_size(size):
    rv = validate._parse_size(size)
    if size == "1 KB":
        assert rv == 1000
    elif size == "1 KiB":
        assert rv == 1024
    elif size == "1 MB":
        assert rv == 1000000
    elif size == "1 MiB":
        assert rv == 1048576
    elif size == "1 GB":
        assert rv == 1000000000
    elif size == "1 GiB":
        assert rv == 1073741824


def test_get_filestorage_size():
    rv = validate._get_filestorage_size(FileStorage(io.BytesIO(b"".ljust(0))))
    assert rv == 0
    rv = validate._get_filestorage_size(FileStorage(io.BytesIO(b"".ljust(123))))
    assert rv == 123
    rv = validate._get_filestorage_size(FileStorage(io.BytesIO(b"".ljust(1024))))
    assert rv == 1024
    rv = validate._get_filestorage_size(FileStorage(io.BytesIO(b"".ljust(1234))))
    assert rv == 1234

    with SpooledTemporaryFile() as temp:
        temp.write(b"".ljust(0))
        rv = validate._get_filestorage_size(FileStorage(temp))
        assert rv == 0

    with SpooledTemporaryFile() as temp:
        temp.write(b"".ljust(123))
        rv = validate._get_filestorage_size(FileStorage(temp))
        assert rv == 123

    with SpooledTemporaryFile() as temp:
        temp.write(b"".ljust(1024))
        rv = validate._get_filestorage_size(FileStorage(temp))
        assert rv == 1024

    with SpooledTemporaryFile() as temp:
        temp.write(b"".ljust(1234))
        rv = validate._get_filestorage_size(FileStorage(temp))
        assert rv == 1234


@pytest.mark.parametrize("size", ["wrong_format", "1.2.3 MiB"])
def test_parse_size_wrong_value(size):
    if size == "wrong_format":
        with pytest.raises(ValueError, match="Invalid size value: "):
            validate._parse_size(size)
    elif size == "1.2.3 MiB":
        with pytest.raises(
            ValueError, match="Invalid float value while parsing size: "
        ):
            validate._parse_size(size)


def test_filesize_min():
    fs = FileStorage(io.BytesIO(b"".ljust(1024)))
    assert validate.FileSize(min="1 KiB", max="2 KiB")(fs) is fs
    assert validate.FileSize(min="0 KiB", max="1 KiB")(fs) is fs
    assert validate.FileSize()(fs) is fs
    assert validate.FileSize(min_inclusive=False, max_inclusive=False)(fs) is fs
    assert validate.FileSize(min="1 KiB", max="1 KiB")(fs) is fs

    with pytest.raises(ValidationError, match="Must be greater than or equal to 2 KiB"):
        validate.FileSize(min="2 KiB", max="3 KiB")(fs)
    with pytest.raises(ValidationError, match="Must be greater than or equal to 2 KiB"):
        validate.FileSize(min="2 KiB")(fs)
    with pytest.raises(ValidationError, match="Must be greater than 1 KiB"):
        validate.FileSize(
            min="1 KiB", max="2 KiB", min_inclusive=False, max_inclusive=True
        )(fs)
    with pytest.raises(ValidationError, match="less than 1 KiB"):
        validate.FileSize(
            min="1 KiB", max="1 KiB", min_inclusive=True, max_inclusive=False
        )(fs)


def test_filesize_max():
    fs = FileStorage(io.BytesIO(b"".ljust(2048)))
    assert validate.FileSize(min="1 KiB", max="2 KiB")(fs) is fs
    assert validate.FileSize(max="2 KiB")(fs) is fs
    assert validate.FileSize()(fs) is fs
    assert validate.FileSize(min_inclusive=False, max_inclusive=False)(fs) is fs
    assert validate.FileSize(min="2 KiB", max="2 KiB")(fs) is fs

    with pytest.raises(ValidationError, match="less than or equal to 1 KiB"):
        validate.FileSize(min="0 KiB", max="1 KiB")(fs)
    with pytest.raises(ValidationError, match="less than or equal to 1 KiB"):
        validate.FileSize(max="1 KiB")(fs)
    with pytest.raises(ValidationError, match="less than 2 KiB"):
        validate.FileSize(
            min="1 KiB", max="2 KiB", min_inclusive=True, max_inclusive=False
        )(fs)
    with pytest.raises(ValidationError, match="greater than 2 KiB"):
        validate.FileSize(
            min="2 KiB", max="2 KiB", min_inclusive=False, max_inclusive=True
        )(fs)


def test_filesize_repr():
    assert (
        repr(
            validate.FileSize(
                min=None, max=None, error=None, min_inclusive=True, max_inclusive=True
            )
        )
        == "<FileSize(min=None, max=None, min_inclusive=True, max_inclusive=True, error=None)>"  # noqa: E501
    )

    assert (
        repr(
            validate.FileSize(
                min="1 KiB",
                max="3 KiB",
                error="foo",
                min_inclusive=False,
                max_inclusive=False,
            )
        )
        == "<FileSize(min='1 KiB', max='3 KiB', min_inclusive=False, max_inclusive=False, error='foo')>"  # noqa: E501
    )


def test_filesize_wrongtype():
    with pytest.raises(TypeError, match="A FileStorage object is required, not "):
        validate.FileSize()(1)


def test_filetype():
    png_fs = FileStorage(io.BytesIO(b"".ljust(1024)), "test.png")
    assert validate.FileType([".png"])(png_fs) is png_fs
    assert validate.FileType([".PNG"])(png_fs) is png_fs

    PNG_fs = FileStorage(io.BytesIO(b"".ljust(1024)), "test.PNG")
    assert validate.FileType([".png"])(PNG_fs) is PNG_fs
    assert validate.FileType([".PNG"])(PNG_fs) is PNG_fs

    with pytest.raises(TypeError, match="A FileStorage object is required, not "):
        validate.FileType([".png"])(1)

    with pytest.raises(
        ValidationError,
        match=r"Not an allowed file type. Allowed file types: \[.*?\]",  # noqa: W605
    ):
        jpg_fs = FileStorage(io.BytesIO(b"".ljust(1024)), "test.jpg")
        validate.FileType([".png"])(jpg_fs)

    with pytest.raises(
        ValidationError,
        match=r"Not an allowed file type. Allowed file types: \[.*?\]",  # noqa: W605
    ):
        no_ext_fs = FileStorage(io.BytesIO(b"".ljust(1024)), "test")
        validate.FileType([".png"])(no_ext_fs)



"""
flask_marshmallow.fields
~~~~~~~~~~~~~~~~~~~~~~~~

Custom, Flask-specific fields.

See the `marshmallow.fields` module for the list of all fields available from the
marshmallow library.
"""

from __future__ import annotations

import re
import typing
from collections.abc import Sequence

from flask import current_app, url_for
from marshmallow import fields, missing

__all__ = [
    "URLFor",
    "UrlFor",
    "AbsoluteURLFor",
    "AbsoluteUrlFor",
    "Hyperlinks",
    "File",
    "Config",
]


_tpl_pattern = re.compile(r"\s*<\s*(\S*)\s*>\s*")


def _tpl(val: str) -> str | None:
    """Return value within ``< >`` if possible, else return ``None``."""
    match = _tpl_pattern.match(val)
    if match:
        return match.groups()[0]
    return None


def _get_value(obj, key, default=missing):
    """Slightly-modified version of marshmallow.utils.get_value.
    If a dot-delimited ``key`` is passed and any attribute in the
    path is `None`, return `None`.
    """
    if "." in key:
        return _get_value_for_keys(obj, key.split("."), default)
    else:
        return _get_value_for_key(obj, key, default)


def _get_value_for_keys(obj, keys, default):
    if len(keys) == 1:
        return _get_value_for_key(obj, keys[0], default)
    else:
        value = _get_value_for_key(obj, keys[0], default)
        # XXX This differs from the marshmallow implementation
        if value is None:
            return None
        return _get_value_for_keys(value, keys[1:], default)


def _get_value_for_key(obj, key, default):
    if not hasattr(obj, "__getitem__"):
        return getattr(obj, key, default)

    try:
        return obj[key]
    except (KeyError, IndexError, TypeError, AttributeError):
        return getattr(obj, key, default)


class URLFor(fields.Field):
    """Field that outputs the URL for an endpoint. Acts identically to
    Flask's ``url_for`` function, except that arguments can be pulled from the
    object to be serialized, and ``**values`` should be passed to the ``values``
    parameter.

    Usage: ::

        url = URLFor("author_get", values=dict(id="<id>"))
        https_url = URLFor(
            "author_get",
            values=dict(id="<id>", _scheme="https", _external=True),
        )

    :param str endpoint: Flask endpoint name.
    :param dict values: Same keyword arguments as Flask's url_for, except string
        arguments enclosed in `< >` will be interpreted as attributes to pull
        from the object.
    :param kwargs: keyword arguments to pass to marshmallow field (e.g. ``required``).
    """

    _CHECK_ATTRIBUTE = False

    def __init__(
        self,
        endpoint: str,
        values: dict[str, typing.Any] | None = None,
        **kwargs,
    ):
        self.endpoint = endpoint
        self.values = values or {}
        fields.Field.__init__(self, **kwargs)

    def _serialize(self, value, key, obj):
        """Output the URL for the endpoint, given the kwargs passed to
        ``__init__``.
        """
        param_values = {}
        for name, attr_tpl in self.values.items():
            attr_name = _tpl(str(attr_tpl))
            if attr_name:
                attribute_value = _get_value(obj, attr_name, default=missing)
                if attribute_value is None:
                    return None
                if attribute_value is not missing:
                    param_values[name] = attribute_value
                else:
                    raise AttributeError(
                        f"{attr_name!r} is not a valid " f"attribute of {obj!r}"
                    )
            else:
                param_values[name] = attr_tpl
        return url_for(self.endpoint, **param_values)


UrlFor = URLFor


class AbsoluteURLFor(URLFor):
    """Field that outputs the absolute URL for an endpoint."""

    def __init__(
        self,
        endpoint: str,
        values: dict[str, typing.Any] | None = None,
        **kwargs,
    ):
        if values:
            values["_external"] = True
        else:
            values = {"_external": True}
        URLFor.__init__(self, endpoint=endpoint, values=values, **kwargs)


AbsoluteUrlFor = AbsoluteURLFor


def _rapply(d: dict | typing.Iterable, func: typing.Callable, *args, **kwargs):
    """Apply a function to all values in a dictionary or
    list of dictionaries, recursively.
    """
    if isinstance(d, (tuple, list)):
        return [_rapply(each, func, *args, **kwargs) for each in d]
    if isinstance(d, dict):
        return {key: _rapply(value, func, *args, **kwargs) for key, value in d.items()}
    else:
        return func(d, *args, **kwargs)


def _url_val(val: typing.Any, key: str, obj: typing.Any, **kwargs):
    """Function applied by `HyperlinksField` to get the correct value in the
    schema.
    """
    if isinstance(val, URLFor):
        return val.serialize(key, obj, **kwargs)
    else:
        return val


class Hyperlinks(fields.Field):
    """Field that outputs a dictionary of hyperlinks,
    given a dictionary schema with :class:`~flask_marshmallow.fields.URLFor`
    objects as values.

    Example: ::

        _links = Hyperlinks(
            {
                "self": URLFor("author", values=dict(id="<id>")),
                "collection": URLFor("author_list"),
            }
        )

    `URLFor` objects can be nested within the dictionary. ::

        _links = Hyperlinks(
            {
                "self": {
                    "href": URLFor("book", values=dict(id="<id>")),
                    "title": "book detail",
                }
            }
        )

    :param dict schema: A dict that maps names to
        :class:`~flask_marshmallow.fields.URLFor` fields.
    """

    _CHECK_ATTRIBUTE = False

    def __init__(self, schema: dict[str, URLFor | str], **kwargs):
        self.schema = schema
        fields.Field.__init__(self, **kwargs)

    def _serialize(self, value, attr, obj):
        return _rapply(self.schema, _url_val, key=attr, obj=obj)


class File(fields.Field):
    """A binary file field for uploaded files.

    Examples: ::

        class ImageSchema(Schema):
            image = File(required=True)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Metadata used by apispec
        self.metadata["type"] = "string"
        self.metadata["format"] = "binary"

    default_error_messages = {"invalid": "Not a valid file."}

    def deserialize(
        self,
        value: typing.Any,
        attr: str | None = None,
        data: typing.Mapping[str, typing.Any] | None = None,
        **kwargs,
    ):
        if isinstance(value, Sequence) and len(value) == 0:
            value = missing
        return super().deserialize(value, attr, data, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        from werkzeug.datastructures import FileStorage

        if not isinstance(value, FileStorage):
            raise self.make_error("invalid")
        return value


class Config(fields.Field):
    """A field for Flask configuration values.

    Examples: ::

        from flask import Flask

        app = Flask(__name__)
        app.config["API_TITLE"] = "Pet API"


        class FooSchema(Schema):
            user = String()
            title = Config("API_TITLE")

    This field should only be used in an output schema. A ``ValueError`` will
    be raised if the config key is not found in the app config.

    :param str key: The key of the configuration value.
    """

    _CHECK_ATTRIBUTE = False

    def __init__(self, key: str, **kwargs):
        fields.Field.__init__(self, **kwargs)
        self.key = key

    def _serialize(self, value, attr, obj, **kwargs):
        if self.key not in current_app.config:
            raise ValueError(f"The key {self.key!r} is not found in the app config.")
        return current_app.config[self.key]



from __future__ import annotations

import typing

import flask
import marshmallow as ma

if typing.TYPE_CHECKING:
    from flask.wrappers import Response


class Schema(ma.Schema):
    """Base serializer with which to define custom serializers.

    See `marshmallow.Schema` for more details about the `Schema` API.
    """

    def jsonify(
        self, obj: typing.Any, many: bool | None = None, *args, **kwargs
    ) -> Response:
        """Return a JSON response containing the serialized data.


        :param obj: Object to serialize.
        :param bool many: Whether `obj` should be serialized as an instance
            or as a collection. If None, defaults to the value of the
            `many` attribute on this Schema.
        :param kwargs: Additional keyword arguments passed to `flask.jsonify`.

        .. versionchanged:: 0.6.0
            Takes the same arguments as `marshmallow.Schema.dump`. Additional
            keyword arguments are passed to `flask.jsonify`.

        .. versionchanged:: 0.6.3
            The `many` argument for this method defaults to the value of
            the `many` attribute on the Schema. Previously, the `many`
            argument of this method defaulted to False, regardless of the
            value of `Schema.many`.
        """
        if many is None:
            many = self.many
        data = self.dump(obj, many=many)
        return flask.jsonify(data, *args, **kwargs)
